%  gtkplot.sl:	utility wrappers for GtkExtra plotting objects {{{
%
%  Copyright (C) 2003-2005 Massachusetts Institute of Technology
% 
%  SLgtk was partially developed at the MIT Center for Space Research,
%  under contract SV1-61010 from the Smithsonian Institution.
% 
%  Permission to use, copy, modify, distribute, and sell this software
%  and its documentation for any purpose is hereby granted without fee,
%  provided that the above copyright notice appear in all copies and
%  that both that copyright notice and this permission notice appear in
%  the supporting documentation, and that the name of the Massachusetts
%  Institute of Technology not be used in advertising or publicity
%  pertaining to distribution of the software without specific, written
%  prior permission.  The Massachusetts Institute of Technology makes
%  no representations about the suitability of this software for any
%  purpose.  It is provided "as is" without express or implied warranty.
%  
%  THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY DISCLAIMS ALL WARRANTIES
%  WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
%  MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL THE MASSACHUSETTS
%  INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
%  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
%  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
%  NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
%  WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
% }}}

% Front matter (type and forward declarations, evalfiles, etc) {{{

#ifexists require
require("gtkextra");
#else
() = evalfile("gtkextra");
#endif

#ifnexists GtkPlotDescriptor
typedef struct {
	plot,			% underlying Gtk widget visualizing the plot(s)
	canvas,			% underlying Gtk canvas on which it's drawn
	width,height,		% display dimensions
	dsets,			% GtkPlotDatasets, one per plot
	dset,			% current plot dataset
	user_data		% for client app to store arbitrary data
} GtkPlotDescriptor;

typedef struct { data, min, max, range, offset} Axis;
typedef struct {
	widget,			% underlying GtkPlotData widget
	axes,			% Axis instances: 0 ==> x Axis, 1 ==> y Axis
	style			% cosmetic specification for this plot
} GtkPlotDataset;

typedef struct {
	sym,
	symstyle,
	symcolor,
	symsize,
	linestyle,
	linecolor,
	vlpsize			% num points a plot can have before being
				% considered a (v)ery (l)arge (p)lot
} GtkPlotStyle;
#endif

define _gtk_plot_set_xrange();
define _gtk_plot_set_yrange();
define _gtk_plot_set_labels();
define _gtk_plot_redraw();
% }}}

% Misc cosmetic styles and preferences {{{

private variable DefaultVeryLargePlotSize = 500000;

private variable LineStyles =
[
	"NONE", "STRAIGHT_LINE", "HISTOGRAM_RIGHT",
	"HISTOGRAM_LEFT", "HISTOGRAM_MIDDLE"
];

private variable Symbols =
[ 
	"NONE", "SQUARE", "CIRCLE",
	"UP_TRIANGLE", "DOWN_TRIANGLE", "RIGHT_TRIANGLE", "LEFT_TRIANGLE",
	"DIAMOND", "PLUS", "CROSS", "STAR", "DOT", "IMPULSE"
];

private variable SymbolSizes =
[
	"0", "1", "2", "3", "4", "5", "6", "7", "8",
	"9", "10", "11", "12", "13", "14", "15", "16"
];

private variable SymbolStyles =
[
  	"EMPTY",
	"FILLED",
	"OPAQUE"
];

define _gtk_plot_linestyles() { return LineStyles; }
define _gtk_plot_symbols() { return Symbols; }
define _gtk_plot_symbolsizes() { return SymbolSizes; }
define _gtk_plot_symbolstyles() { return SymbolStyles; }

private variable DefaultPlotStyle = NULL;

private define get_default_plot_style()
{
   if (DefaultPlotStyle == NULL) {
	variable style = @GtkPlotStyle;
	% default: smallest symbol, with empty interior, for fastest rendering
	style.sym = GTK_PLOT_SYMBOL_DOT;
   	style.symstyle = GTK_PLOT_SYMBOL_EMPTY;
   	style.symcolor = @gdk_black;
   	style.symsize  = 4;
   	style.linestyle  = GTK_PLOT_CONNECT_STRAIGHT;
   	style.linecolor  = @gdk_black;
	style.vlpsize	 = DefaultVeryLargePlotSize;
   	DefaultPlotStyle = style;
   }
   return DefaultPlotStyle;
}

define _gtk_plot_set_vlpsize()
{
   variable vlpsize = NULL;
   switch (_NARGS)
   { case 0: vlpsize = DefaultVeryLargePlotSize; }
   { case 1: vlpsize = int(()); }

   if (orelse {vlpsize == NULL} {typeof(vlpsize) != Integer_Type})
	usage(_function_name+"([Integer_Type new_very_large_plot_size_value])");

   get_default_plot_style().vlpsize = vlpsize;
}

private define set_style_symbol()
{
   variable symtype, symsize = NULL, symcolor = NULL, symstyle = NULL;
   variable style, usg;
   (style,usg) = ();

   switch (_NARGS - 2)
	{ case 1: symtype = (); }
	{ case 2: (symtype,symcolor) = (); }
	{ case 3: (symtype,symcolor,symsize) = (); }
	{ case 4: (symtype,symcolor,symsize,symstyle) = (); }
	{ usage( usg); }

   style.sym = symtype;
   if (symsize  != NULL) style.symsize  = symsize;
   if (symstyle != NULL) style.symstyle = symstyle;

   % Equality test used to validate that input color is indeed a GdkColor struct
   if (andelse	{symcolor != NULL}
		{not(gdk_color_equal(style.symcolor,symcolor))})
	style.symcolor = @symcolor;
}

define _gtk_plot_style_pane_new(dset)
{
   % WIPAPU = work in progress, and purposely undocumented
   variable style_pane = struct {
	sym,
	symstyle,
	symcolor,
	symsize,

	linestyle,

	widget
   };

   variable sp = @style_pane;
   variable style = dset.style;

   variable vbox = gtk_vbox_new(TRUE,5);
   sp.linestyle = _pref_new("Line style      ",LineStyles,style.linestyle);
   gtk_container_add(vbox,sp.linestyle.widget);

   sp.sym = _pref_new("Symbol", Symbols, style.sym);
   gtk_container_add(vbox,sp.sym.widget);

   sp.symcolor = _pref_new_from_func ("Symbol color ", style.symcolor,
   			&gdk_color_equal, &_color_button_new);
   gtk_container_add(vbox,sp.symcolor.widget);

   sp.symstyle = _pref_new("Symbol style", SymbolStyles, style.symstyle);
   gtk_container_add(vbox,sp.symstyle.widget);

   sp.symsize = _pref_new("Symbol size",SymbolSizes, style.symsize);
   gtk_container_add(vbox,sp.symsize.widget);

   sp.widget = vbox;

   return sp;
}

private define set_axis_label_style(plot,axis,range,angle,fgcolor)
{
   if (range != NULL) {
	if (orelse {range >= 10000} {range <= 0.001} )
	   gtk_plot_axis_set_labels_style(plot, axis, GTK_PLOT_LABEL_POW,1);
	else
	   gtk_plot_axis_set_labels_style(plot, axis, GTK_PLOT_LABEL_FLOAT,1);
   }

   gtk_plot_axis_set_labels_attributes(plot, axis, NULL, 10, angle,
		 		fgcolor, NULL, TRUE, GTK_JUSTIFY_CENTER);
}

% }}}

% Private canvas/plot support routines (creation, defaults, etc)  {{{

% Axes (creation, ranges, etc) {{{
private define round_number(x)
{
   if (x == 0.0) return 0.0;	% adapted from algorithm in PgPlot library
   
   variable xx = abs(x);
   variable xlog = log10(xx);
   variable ilog = int(xlog);
   if (xlog < 0.)
      ilog -= 1;

   variable pwr = 10.0 ^ ilog;
   variable nice, nsub, frac = xx/pwr;

   if (frac <= 2.0)
      nice = 2.0;
   else if (frac <= 5.0)
      nice = 5.0;
   else
      nice = 10.0;

   if (x < 0.0)
	return (-pwr * nice);

   return  (pwr * nice);
}

#ifeval _slang_version >= 20000
private define bad(x) { orelse {isinf(x)} {isnan(x)}; }
#endif

private define axis_new(plotd, which_axis, data)
{
   variable ax = @Axis;
   ax.data = data;
   ax.min = min(data);
   ax.max = max(data);
   ax.range = ax.max - ax.min;

#ifeval _slang_version >= 20000
   % This test is not foolproof, but does catch most major problems while
   % avoiding 2 more isinf()/isnan() traversals of both vectors (slow!)
   if (orelse {bad(ax.min)} {bad(ax.max)} {bad(ax.range)})
	error("Input data contains Inf/Nan; cull before plotting");
#endif

   if (ax.range == 0) { ax.range = 0.5; ax.max += ax.range; }

   % Would like a prettier plot by adding an offset between the data
   % and axes, but the GtkPlot ticks code is fragile/unpredictable
   % when the x/y range is fudged and log scaling is invoked.
   ax.offset = ax.range / 50.0;

   if (length(plotd.dsets) == 0) {
	variable Min, Max;
	if (which_axis == 0) {
	   gtk_plot_get_xrange(plotd.plot,&Min,&Max);
	   if (orelse {ax.min < Min} {ax.max > Max})
		_gtk_plot_set_xrange(plotd, ax.min , ax.max );
	}
	else {
	   gtk_plot_get_yrange(plotd.plot,&Min,&Max);
	   if (orelse {ax.min < Min} {ax.max > Max})
		_gtk_plot_set_yrange(plotd, ax.min , ax.max );
	}
   }

   return ax;
}
% }}}
private define add_plot_dataset()
{
   _stk_reverse(_NARGS);

   variable pd, x, y, usg = ();
   variable style = (@get_default_plot_style)();
   variable lcolor = style.linecolor, lstyle = style.linestyle;

   switch(_NARGS-1)
      {case 3: ( y, x, pd) = (); }
      {case 4: ( lcolor, y, x, pd ) = (); style.symcolor = lcolor; }
      {case 5: ( lstyle, lcolor, y, x, pd) = ();  style.symcolor = lcolor; }
      { usage(usg); }

   if (typeof(pd) != GtkPlotDescriptor)
	usage(usg + "\nMissing GtkPlotDescriptor argument");

   variable len = length(x);
   if (len != length(y))
	usage(usg + "\nInput must contain 2 numeric vectors, of"+
					" equal but nonzero length");

   x = typecast(x,Double_Type);		% these will throw errors when
   y = typecast(y,Double_Type);		% given inappropriate input

   if (len > style.vlpsize) {
	lstyle = GTK_PLOT_CONNECT_NONE;
	if (style.sym == GTK_PLOT_SYMBOL_NONE)
	   style.sym = GTK_PLOT_SYMBOL_DOT;
   }

   variable ds = @GtkPlotDataset;
   ds.widget = gtk_plot_data_new();
   if (ds.widget == NULL) error("could not create new plot dataset");

   gtk_plot_data_set_x(ds.widget, x);
   gtk_plot_data_set_y(ds.widget, y);
   gtk_plot_data_set_numpoints(ds.widget,len);
   gtk_plot_add_data(pd.plot,ds.widget);

   gtk_plot_data_set_symbol(ds.widget, style.sym , style.symstyle,
			style.symsize, 0, style.symcolor, style.symcolor);
   gtk_plot_data_set_line_attributes(ds.widget, GTK_PLOT_LINE_SOLID,
   							0, 0, 1, lcolor);
   gtk_plot_data_set_connector(ds.widget, lstyle);
   _gtk_plot_set_labels(pd, "", "");

   ds.style = style;
   ds.axes  = [	axis_new(pd, 0, x), axis_new(pd, 1, y) ];
   pd.dsets = [ pd.dsets, ds ];
   pd.dset  = ds;
}

private define canvas_delete_item(canvas,child) { return TRUE; }

private define canvas_size_allocate(canvas,allocation,pd)
{
   if (orelse {pd.width != allocation.width} {pd.height != allocation.height}) {
	% BUG HACK-AROUND: check new size against previous, to prevent inf loop
	pd.width = allocation.width;
	pd.height = allocation.height;
	gtk_plot_canvas_set_size(canvas,pd.width,pd.height);
   }
   else
	_gtk_plot_redraw(pd);
}

private define set_common_defaults(pd)
{
   gtk_plot_canvas_set_flags(pd.canvas,  GTK_PLOT_CANVAS_CAN_DND |
		GTK_PLOT_CANVAS_CAN_SELECT | GTK_PLOT_CANVAS_CAN_SELECT_ITEM);

   % Default to caller having ability to delete objects from canvas
   () = g_signal_connect(pd.canvas,"delete_item",&canvas_delete_item);

   % Allow canvas to automatically resize
   () = g_signal_connect_after(pd.canvas, "size_allocate",
			       			&canvas_size_allocate, pd);

   gtk_plot_hide_legends(pd.plot);
   gtk_plot_clip_data(pd.plot, TRUE);
   gtk_plot_axis_hide_title(pd.plot, GTK_PLOT_AXIS_TOP);
   gtk_plot_axis_hide_title(pd.plot, GTK_PLOT_AXIS_RIGHT);
   gtk_plot_axis_show_labels(pd.plot, GTK_PLOT_AXIS_TOP, FALSE);
   gtk_plot_axis_show_labels(pd.plot, GTK_PLOT_AXIS_RIGHT, FALSE);
}

private define create_canvas_and_plot(pd)
{
   pd.canvas  = gtk_plot_canvas_new( int(GTK_PLOT_LETTER_W * 0.85),
					int(GTK_PLOT_LETTER_H * 0.70), 1.0);
   if (pd.canvas == NULL) error("could not create new plot canvas");

   pd.plot = gtk_plot_new_with_size(NULL, 0.8, 0.8);
   if (pd.plot == NULL) error("could not create new plot");
   gtk_plot_canvas_add_plot(pd.canvas, pd.plot, 0.12, 0.10);
}

private define set_plot_background_image()
{
   variable dims, dim, pd, image, pixbuf, usg = (), reflect;

   switch(_NARGS - 1)
      { case 2: (image, pd) = (); reflect = 0;}
      { case 3: (image, reflect, pd) = (); }
      { usage(usg); }

   !if ( length(where(_gdk_query_depths() == 24)))
	usage("image plot requires 24-bit RGB display capability");

   switch (typeof(image))
   { case Array_Type:

	(dims, dim, ) = array_info(image);
	switch(dim)
	{ case 1:
	   	variable size = sqrt(dims[0]);
		if (int(size) != size)
		   error("1D pixel array does not represent an NxN image");
		size = int(size);
		dims = [size, size];
	}
	{ case 2: }
	{ error("input pixel array must be 1D or 2D"); }

      	image = norm_array(image);
      	reshape(image, dims);

	if (reflect)
	   image = image [ [ dims[0] - 1 : 0: -1 ], * ];

	variable rgbbuf = UChar_Type[ dims[0], dims[1], 3];
	rgbbuf[*,*,0] = image;
	rgbbuf[*,*,1] = image;
	rgbbuf[*,*,2] = image;

	pixbuf = gdk_pixbuf_new_from_data(rgbbuf);
	rgbbuf = NULL;
   }
   { case GdkPixbuf:

	pixbuf = image; 
	dims = [ gdk_pixbuf_get_height(pixbuf),
		   			gdk_pixbuf_get_width(pixbuf)];
   }
   { usage(usg); }

   variable pixmap = gdk_pixmap_new(NULL, dims[1], dims[0], 24);
   gdk_draw_pixbuf(pixmap,NULL,pixbuf,0,0,0,0,-1,-1,GDK_RGB_DITHER_NONE,0,0);
   pixbuf = NULL;

   % Shrink canavas to size of image, plus a small border
   gtk_plot_canvas_set_size(pd.canvas, dims[1]+10, dims[0]+10);

   gtk_plot_set_background_pixmap(pd.plot,pixmap);

   _gtk_plot_set_yrange(pd, 0, dims[0]);
   _gtk_plot_set_xrange(pd, 0, dims[1]);

   %g_object_unref(pixmap);
}

private define print_cb(pc, plotd)
{
   variable fname;
   if (pc.dest == 0)
	fname = _tmp_file_name("gtkplot");
   else
	fname = pc.fname;

   % Generate an encapsulated postscript file
   () = gtk_plot_canvas_export_ps(plotd.canvas,fname, pc.orient, TRUE, pc.size);

   if (pc.dest == 0) {
	() = system( sprintf("%s %s",pc.cmd, fname));
	() = remove(fname);
   }
}

private define invoke_print_cb()
{
   variable window, invoke_dialog,ctx, plotd;
   (window, invoke_dialog, ctx, plotd) = ();

   if (invoke_dialog)
	_print_dialog(ctx, &print_cb, plotd);
   else
	print_cb(ctx, plotd);

   if (window != NULL) {
	gtk_widget_destroy(window);
	gtk_main_quit();
   }

   return FALSE;
}
% }}}

% Main Public interface {{{
define _gtk_plot_set_xrange(plotd,xmin,xmax)
{
   variable range = xmax - xmin;
   variable step = round_number( range / 8.0);
   gtk_plot_set_xrange(plotd.plot, xmin , xmax );
   gtk_plot_axis_set_ticks(plotd.plot, GTK_PLOT_AXIS_X, step, 2);
   set_axis_label_style(plotd.plot, GTK_PLOT_AXIS_BOTTOM, range, 0, NULL);
}

define _gtk_plot_set_yrange(plotd,ymin,ymax)
{
   variable range = (ymax - ymin);
   variable step = round_number( range / 5.0);
   gtk_plot_set_yrange(plotd.plot, ymin , ymax );
   gtk_plot_axis_set_ticks(plotd.plot, GTK_PLOT_AXIS_Y, step, 2);
   set_axis_label_style(plotd.plot, GTK_PLOT_AXIS_LEFT, range, 90, NULL);
}

define _gtk_plot_set_xlabel(plotd,label)
{
   gtk_plot_axis_set_title(plotd.plot,GTK_PLOT_AXIS_BOTTOM, label);
}

define _gtk_plot_set_ylabel(plotd,label)
{
   gtk_plot_axis_set_title(plotd.plot,GTK_PLOT_AXIS_LEFT, label);
}

define _gtk_plot_set_labels()
{
   variable plotd, xlabel, ylabel, xcolor = NULL, ycolor = NULL;
   switch (_NARGS)
	{  case 3: (plotd, xlabel, ylabel) = (); }
	{  case 4: (plotd, xlabel, ylabel, xcolor) = (); ycolor = xcolor;}
	{  case 5: (plotd, xlabel, ylabel, xcolor, ycolor) = (); }

   gtk_plot_axis_set_title(plotd.plot,GTK_PLOT_AXIS_BOTTOM, xlabel);
   set_axis_label_style(plotd.plot, GTK_PLOT_AXIS_BOTTOM, NULL, 0, xcolor);

   gtk_plot_axis_set_title(plotd.plot,GTK_PLOT_AXIS_LEFT, ylabel);
   set_axis_label_style(plotd.plot, GTK_PLOT_AXIS_LEFT, NULL, 90, ycolor);
}

% _gtk_plot_set_symbol_*: WIPAPU
define _gtk_plot_set_symbol_defaults()
{
   variable usg = sprintf("Void_Type  %s(type, "+
			"[ gdkcolor [, size [, style ]]])", _function_name);
   variable args = __pop_args(_NARGS);
   set_style_symbol(  __push_args(args), get_default_plot_style(), usg);
}

define _gtk_plot_set_symbol()
{
   variable usg = sprintf("Void_Type  %s(GtkPlotDataset, type, "+
			"[ gdkcolor, [, size [, style ]]])", _function_name);

   if (_NARGS < 2)
	usage(usg);

   variable args = __pop_args(_NARGS - 1);
   variable dataset = (), style = dataset.style;

   set_style_symbol( __push_args(args), style, usg);
   gtk_plot_data_set_symbol(dataset.widget, style.sym , style.symstyle,
			style.symsize, 0, style.symcolor, style.symcolor);
}

define _gtk_plot()
{
   variable args = __pop_args(_NARGS);
   variable pd = @GtkPlotDescriptor;
   variable u = "GtkPlotDescriptor = gtk_plot(x, y [,linecolor [,linestyle]])";

   create_canvas_and_plot(pd);
   pd.dsets = GtkPlotDataset[0];
   add_plot_dataset(u, pd, __push_args(args) );
   set_common_defaults(pd);

   return pd;
}

define _gtk_oplot()
{
   variable args = __pop_args(_NARGS);
   variable u = "Void_Type  gtk_oplot(GtkPlotDescriptor, x, y "+
   						"[,linecolor [,linestyle]] )";
   add_plot_dataset(u, __push_args(args) );
}

define _gtk_plot_image()
{
   variable args = __pop_args(_NARGS);
   variable pd = @GtkPlotDescriptor;
   variable u = "GtkPlotDescriptor = _gtk_plot_image("+
		"image_pixel_array | GdkPixbuf [, reflect])";

   create_canvas_and_plot(pd);
   pd.dsets = GtkPlotDataset[0];
   set_plot_background_image(__push_args(args), pd, u);
   _gtk_plot_set_labels(pd, "", "");

   gtk_plot_axis_show_ticks(pd.plot, GTK_PLOT_AXIS_TOP,FALSE,FALSE);
   gtk_plot_axis_show_ticks(pd.plot, GTK_PLOT_AXIS_RIGHT, FALSE, FALSE);

   set_common_defaults(pd);
   return pd;
}

define _gtk_plot_remove(plotd,index)
{
   variable nplots = length(plotd.dsets);
   if (orelse {index < 1} {index > nplots})	% be nice to humanity, by
      return;					% letting index start at 1

   index--;
   variable dataset = plotd.dsets[index];

   % Observe that we intentionally avoid redrawing
   () = gtk_plot_remove_data(plotd.plot, dataset.widget);

   variable indices = Integer_Type[nplots] + 1;
   indices[index] = 0;
   plotd.dsets = plotd.dsets[where(indices)];

   % Reposition the current dataset pointer
   nplots--;
   switch (nplots)
	{ case 0: plotd.dset = NULL; }		 % no more datasets left!
	{ if (__eqs(plotd.dset,dataset)) {	 % otherwise, if we just
		if (index == nplots)		 % deleted the current ds
		   index--;			 % then reset the current
		plotd.dset = plotd.dsets[index]; % ds pointer to the youngest
	  }				 	 % adjacent dataset
	}
}

define _gtk_plot_redraw()
{
   if (_NARGS == 1) {
	variable plotd = ();
	gtk_plot_canvas_paint(plotd.canvas);
	gtk_plot_canvas_refresh(plotd.canvas);
	return;
   }
   usage("Void_Type _gtk_plot_redraw(GtkPlotDescriptor);");
}

define _gtk_plot_transform(plotd,x,y,direction)
{
   % Transform a point from dataset/canvas space to canvas/dataset space
   switch (direction)
   { case 1:
	gtk_plot_canvas_get_pixel(plotd.canvas,x,y,&x,&y);  % canvas to data
	gtk_plot_get_point(plotd.plot,x,y,&x,&y);
   }
   { case -1:
	gtk_plot_get_pixel(plotd.plot,x,y,&x,&y);	    % data to canvas
	gtk_plot_canvas_get_position(plotd.canvas, int(x), int(y), &x, &y);
   }
   return (x, y);
}

define _gtk_plot_print()
{
   variable plotd = NULL, ctx, invoke_dialog = 0;
   variable u = "gtk_plot_print (GtkPlotDescriptor "+
			"[, invoke_dialog_flag | GtkPrintContext] );";

   switch(_NARGS)
   { case 1: plotd = (); ctx = _print_context_new(); ctx.dest = 1; }
   { case 2: (plotd, ctx) = (); }

   % Be extra careful, to avoid a hanging gtk_main() from bad input
   if (orelse {plotd == NULL} {typeof(plotd) != GtkPlotDescriptor})
	usage(u);

   switch (typeof(ctx))
   { case GtkPrintContext: ctx = _print_context_set_defaults(ctx); }
   { case Integer_Type: invoke_dialog = ctx; ctx = _print_context_new(); }
   { usage(u); }

   if (gtk_main_level() == 0) {
	variable w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_add(w,plotd.canvas);
	gtk_widget_show_all(w);
	ctx.dest = 1;			% default output to file
	() = gtk_idle_add(&invoke_print_cb, w, invoke_dialog, ctx, plotd);
	gtk_main();
   }
   else 
	() = invoke_print_cb(NULL, invoke_dialog, ctx, plotd);
}
% }}}

gtk_plot_axis_default_width(0.25);
gtk_plot_axis_default_tick_length(4);
gtk_plot_axis_default_tick_width(0.25);

#ifexists provide
provide("gtkplot");
#endif
